/* dynamic.c - contains code for the ELFdump functions that parse
 * the dynamic linking information */
/* Written by Owen Klan  -  October 2002 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <elf.h>

#include "elfdump.h"

/* Local function prototypes */
static char *get_dyn_tagname(Elf32_Dyn *dyn);

/* Procedure to display dynamic linker information in a column
 * formatted display */
void dynamic_report(Elf32_Ehdr *header, char *image)
{
    Elf32_Off sec_offset;	       /* File offset to section table */
    Elf32_Half sec_size;	       /* Size of section header entries */
    Elf32_Half num_secs;	       /* Number of sections in file */
    Elf32_Shdr *sec_ptr;	       /* Pointer to current section entry */
    Elf32_Dyn *dyn_ptr;		       /* Pointer to current Dynamic entry */
    int i = 0;
    
    if (!header || !image)
      return;

    sec_size = ehalf_swap(header->e_shentsize);
    num_secs = ehalf_swap(header->e_shnum);
    sec_offset = eword_swap(header->e_shoff);
    
    /* Find the dynamic table in the sections table */
    sec_ptr = (Elf32_Shdr *)(image + sec_offset);
    for (i = 0; i < num_secs; i++) {
	if (eword_swap(sec_ptr->sh_type) == SHT_DYNAMIC) {
	    i = 0;
	    break;
	} else {
	    sec_ptr++;
	    continue;
	}
    }
    
    dyn_ptr = (Elf32_Dyn *)(image + eword_swap(sec_ptr->sh_offset));
    
    if (i != 0) {
	fprintf(stdout, "Could not find a dynamic linker section! This is"
		" most likely a staticly-linked \nbinary or a relocatable"
		" file.\n");
	return;
    }

    fprintf(stdout, "\nDynamic Linkage Information\n");
    fprintf(stdout, "Entry  %-12s  Value\n", "Tag Value");
    
    /* We have actually found the .dynamic section, begin parsing the
     * entries */
    do {
	char *tag_type = NULL;
	
	tag_type = get_dyn_tagname(dyn_ptr);
	fprintf(stdout, " %3d   %-12s  %08X\n", i, tag_type,
		eword_swap(dyn_ptr->d_un.d_val));
	i++;
	dyn_ptr++;
    } while (eword_swap(dyn_ptr->d_tag) != DT_NULL);
}

/* Procedure that displays information regarding dynamic linking
 * symbols */
void dynsym_report(Elf32_Ehdr *header, char *image)
{
    char *dynstr_tab;		       /* Pointer to the string table */
    Elf32_Off sec_offset;	       /* File offset to section table */
    Elf32_Half sec_size;	       /* Size of section header entries */
    Elf32_Half num_secs;	       /* Number of sections in file */
    Elf32_Shdr *sec_ptr;	       /* Pointer to current section entry */
    Elf32_Sym *sym_ptr;		       /* Pointer to current Dynamic symbol */
    int i = 0;
    int num_syms;
    char *type_str = NULL, *bind_str = NULL; 
    char slink_str[8];
    
    if (!header || !image)
      return;

    sec_size = ehalf_swap(header->e_shentsize);
    num_secs = ehalf_swap(header->e_shnum);
    sec_offset = eword_swap(header->e_shoff);
    
    /* Find the dynamic string table in the sections table */
    sec_ptr = (Elf32_Shdr *)(image + sec_offset);
    for (i = 0; i < num_secs; i++) {
	if (strncmp((name_table + eword_swap(sec_ptr->sh_name)),
		    ".dynstr", 7) == 0) {
	    i = 0;
	    break;
	} else {
	    sec_ptr++;
	    continue;
	}
    }
    
    if (i != 0) {
	fprintf(stdout, "Cannot find dynamic section's string table!\n");
	return;
    }

    dynstr_tab = (char *)(image + eword_swap(sec_ptr->sh_offset));
    
    /* Now find the .dynsym section */
    num_secs = ehalf_swap(header->e_shnum);
    sec_offset = eword_swap(header->e_shoff);
   
    /* Find the dynamic symbol table in the sections table */
    sec_ptr = (Elf32_Shdr *)(image + sec_offset);
    for (i = 0; i < num_secs; i++) {
	if (eword_swap(sec_ptr->sh_type) == SHT_DYNSYM) {
	    i = 0;
	    break;
	} else {
	    sec_ptr++;
	    continue;
	}
    }

    if (i != 0) {
	fprintf(stdout, "No dynamic symbol table found! This is either a "
		"statically-linked binary or\n");
	fprintf(stdout, "a relocatable object file!\n");
	return;
    }
    
    num_syms = eword_swap(sec_ptr->sh_size) / eword_swap(sec_ptr->sh_entsize);
    sym_ptr = (Elf32_Sym *)(image + eword_swap(sec_ptr->sh_offset));
    
    fprintf(stdout, "\nDynamic Linkage Symbol Information\n");
    fprintf(stdout, "Symbol  Value     %6s  Type  Binding  %8s  "
	    "Name\n", "Size", "Sec.Link");
    for (i=0; i<num_syms; i++) {
	type_str = get_sym_type(sym_ptr);
	bind_str = get_sym_bind(sym_ptr);
	sprintf(slink_str, "%d", ehalf_swap(sym_ptr->st_shndx));
	
	fprintf(stdout, " %4d   %08X  %6X  %s   %3s       %2s      %s\n", i,
		eword_swap(sym_ptr->st_value), eword_swap(sym_ptr->st_size),
		type_str, bind_str, slink_str, 
		(dynstr_tab + eword_swap(sym_ptr->st_name)));
	sym_ptr++;
	if (type_str)
	  free(type_str);
	if (bind_str)
	  free(bind_str);
    }
}

/* Procedure that displays shared object dependancies for an ELF
 * file */
void display_depends(Elf32_Ehdr *header, char *image)
{
    char *dynstr_tab;		       /* Pointer to the string table */
    Elf32_Off sec_offset;	       /* File offset to section table */
    Elf32_Half sec_size;	       /* Size of section header entries */
    Elf32_Half num_secs;	       /* Number of sections in file */
    Elf32_Shdr *sec_ptr;	       /* Pointer to current section entry */
    Elf32_Dyn *dyn_ptr;		       /* Pointer to current Dynamic entry */
    int i = 0;
    
    if (!header || !image)
      return;

    sec_size = ehalf_swap(header->e_shentsize);
    num_secs = ehalf_swap(header->e_shnum);
    sec_offset = eword_swap(header->e_shoff);
    
    /* Find the dynamic string table in the sections table */
    sec_ptr = (Elf32_Shdr *)(image + sec_offset);
    for (i = 0; i < num_secs; i++) {
	if (strncmp((name_table + eword_swap(sec_ptr->sh_name)),
		    ".dynstr", 7) == 0) {
	    i = 0;
	    break;
	} else {
	    sec_ptr++;
	    continue;
	}
    }
    
    if (i != 0) {
	fprintf(stdout, "Cannot find dynamic section's string table!\n");
	return;
    }

    dynstr_tab = (char *)(image + eword_swap(sec_ptr->sh_offset));

    
    /* Now find the .dynamic section again */
    num_secs = ehalf_swap(header->e_shnum);
    sec_offset = eword_swap(header->e_shoff);
   
    /* Find the dynamic table in the sections table */
    sec_ptr = (Elf32_Shdr *)(image + sec_offset);
    for (i = 0; i < num_secs; i++) {
	if (eword_swap(sec_ptr->sh_type) == SHT_DYNAMIC) {
	    i = 0;
	    break;
	} else {
	    sec_ptr++;
	    continue;
	}
    }
    
    fprintf(stdout, "\nShared Object Dependencies:\n");

    /* Display the contents of the DT_NEEDED field */
    dyn_ptr = (Elf32_Dyn *)(image + eword_swap(sec_ptr->sh_offset));
    do {
	if (eword_swap(dyn_ptr->d_tag) == DT_NEEDED) {
	    fprintf(stdout, " %s\n",
		    (dynstr_tab + eword_swap(dyn_ptr->d_un.d_val)));
     	} else {
	    dyn_ptr++;
	    continue;
	}
	dyn_ptr++;
    } while (eword_swap(dyn_ptr->d_tag) != DT_NULL);
}

/* Function that will generate an ASCII string representing a
 * Dynamic section entry's d_tag field */
static char *get_dyn_tagname(Elf32_Dyn *dyn)
{
    char *tag_str = NULL;
    int tag_type;
    
    if (!dyn)
      return NULL;
    
    if ((tag_str = malloc(13)) == NULL)
      return NULL;
    else
      memset(tag_str, 0x00, 13);
    
    tag_type = eword_swap(dyn->d_tag);
    
#if defined(DT_LOPROC) && defined(DT_HIPROC)
    if (tag_type >= DT_LOPROC || tag_type <= DT_HIPROC)
      sprintf(tag_str, "CPU_SPECIFIC");
#endif

    switch (tag_type) {
    case DT_NULL:     sprintf(tag_str, "DT_NULL"); break;
    case DT_NEEDED:   sprintf(tag_str, "DT_NEEDED"); break;
    case DT_PLTRELSZ: sprintf(tag_str, "DT_PLTRELSZ"); break;
    case DT_PLTGOT:   sprintf(tag_str, "DT_PLTGOT"); break;
    case DT_HASH:     sprintf(tag_str, "DT_HASH"); break;
    case DT_STRTAB:   sprintf(tag_str, "DT_STRTAB"); break;
    case DT_SYMTAB:   sprintf(tag_str, "DT_SYMTAB"); break;
    case DT_RELA:     sprintf(tag_str, "DT_RELA"); break;
    case DT_RELASZ:   sprintf(tag_str, "DT_RELASZ"); break;
    case DT_RELAENT:  sprintf(tag_str, "DT_RELAENT"); break;
    case DT_STRSZ:    sprintf(tag_str, "DT_STRSZ"); break;
    case DT_SYMENT:   sprintf(tag_str, "DT_SYMENT"); break;
    case DT_INIT:     sprintf(tag_str, "DT_INIT"); break;
    case DT_FINI:     sprintf(tag_str, "DT_FINI"); break;
    case DT_SONAME:   sprintf(tag_str, "DT_SONAME"); break;
    case DT_RPATH:    sprintf(tag_str, "DT_RPATH"); break;
    case DT_SYMBOLIC: sprintf(tag_str, "DT_SYMBOLIC"); break;
    case DT_REL:      sprintf(tag_str, "DT_REL"); break;
    case DT_RELSZ:    sprintf(tag_str, "DT_RELSZ"); break;
    case DT_RELENT:   sprintf(tag_str, "DT_RELENT"); break;
    case DT_PLTREL:   sprintf(tag_str, "DT_PLTREL"); break;
    case DT_DEBUG:    sprintf(tag_str, "DT_DEBUG"); break;
    case DT_TEXTREL:  sprintf(tag_str, "DT_TEXTREL"); break;
    case DT_JMPREL:   sprintf(tag_str, "DT_JMPREL"); break;
    }

    return tag_str;
}

/* Function that returns the value of a dynamic symbol. */
Elf32_Word get_dynsym_value(Elf32_Ehdr *header, char *image, int n)
{
    Elf32_Shdr *sec;
    Elf32_Sym *sym;
    int i;
    int num_syms;
    int found = 0;
    
    sec = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    for (i = 0; i < eword_swap(header->e_shnum); i++, sec++) {
	if (eword_swap(sec->sh_type) == SHT_DYNSYM) {
	    found = 1;
	    break;
	}
    }
    
    if (!found) {
	fprintf(stdout, "Couldn't find dynamic linkage symbol table!\n");
	return -1;
    }
    
    num_syms = eword_swap(sec->sh_size) / eword_swap(sec->sh_entsize);
    if (n > num_syms || n < 0) {
	fprintf(stdout, "Bad symbol value: %d\n", n);
	return -1;
    }

    sym = (Elf32_Sym *)(image + eword_swap(sec->sh_offset));
    sym += n;
    
    return eword_swap(sym->st_value);
}
