/* sections.c - contains all of the section related
 * functions for elfdump.
 * Written by Owen Klan  -  August 2002 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <elf.h>

#include "elfdump.h"

char *get_sec_type(Elf32_Shdr *header);
char *get_sec_flags(Elf32_Shdr *header);
void dump_sec(Elf32_Shdr *header, int n);

/* External definition for the name string table */
extern char *name_table;

/* Reads and displays the contents of the sections table. */
void display_sections(Elf32_Ehdr *header, char *image)
{
    Elf32_Off sec_offset;	       /* section table offset */
    Elf32_Half sec_size;      	       /* size of section table */
    Elf32_Half num_secs;	       /* num. elements in section table */
    int i;
    int num_unused = 0;		       /* number of unused section entries */
    Elf32_Shdr *sec_table, *ptr;       /* section table buffer */
    
    sec_offset = eword_swap(header->e_shoff);
    sec_size = ehalf_swap(header->e_shentsize) * ehalf_swap(header->e_shnum);
    num_secs = ehalf_swap(header->e_shnum);

    fprintf(stdout, "\nSection Details\n");

    /* 
     * point to the section table in the ELF image
     */
    sec_table = (Elf32_Shdr *)(image + sec_offset);
    ptr = (Elf32_Shdr *)sec_table;     /* ptr is an increment pointer here.. */
    
    /* process sections, one at a time */
    for (i=0;i<num_secs;i++) {
	if (eword_swap(ptr->sh_type) != SHT_NULL) {
	    dump_sec(ptr, i);
	}
	else
	  num_unused++;
	
	ptr++;
    }
    
    if (num_unused > 0)
      fprintf(stdout, "%d unused sections were not displayed\n",
	      num_unused);
}

/* Procedure that displays section information in rather brief columns */
void sections_report(Elf32_Ehdr *header, char *image)
{
    Elf32_Off sec_offset;	       /* section table offset */
    Elf32_Half sec_size;      	       /* size of section table */
    Elf32_Half num_secs;	       /* num. elements in section table */
    int i;
    Elf32_Shdr *sec_table, *ptr;       /* section table buffer */
    
    sec_offset = eword_swap(header->e_shoff);
    sec_size = ehalf_swap(header->e_shentsize) * ehalf_swap(header->e_shnum);
    num_secs = ehalf_swap(header->e_shnum);
    fprintf(stdout, "\nSection Details\n");
    /* 
     * point to the section table in the ELF image
     */
    sec_table = (Elf32_Shdr *)(image + sec_offset);
    ptr = (Elf32_Shdr *)sec_table;     /* ptr is an increment pointer here.. */
    
    fprintf(stdout, "%-7s  VirtAddr  FileOff.  %8s  %-20s  Alignment\n",
	    "Section", "Size", "Name");
    /* process sections, one at a time */
    for (i=0;i<num_secs;i++) {
	fprintf(stdout, " %3d     %08X  %08X  %8X  %-20s    %d\n",
		i, eword_swap(ptr->sh_addr), eword_swap(ptr->sh_offset),
		eword_swap(ptr->sh_size),
		(name_table + eword_swap(ptr->sh_name)),
		eword_swap(ptr->sh_addralign));
	ptr++;
    }
}


/* Procedure to display section information to
 * the standard output */
void dump_sec(Elf32_Shdr *header, int n)
{
    char *flags_str = NULL;
    char *type_str = NULL;
    
    /* don't attempt to dereference a NULL pointer */
     if (header == NULL)
      return;

    flags_str = get_sec_flags(header);
    type_str = get_sec_type(header);
    
    fprintf(stdout, "\nSection %d\t\tName:  %s\n", n,
	    (name_table + eword_swap(header->sh_name)));
    fprintf(stdout, " Alignment:  %d\t\tType:  %s\n",
	    eword_swap(header->sh_addralign), type_str);
    fprintf(stdout, " Execution Virtual Address:  0x%08X\t\t Size: %d\n",
	    eword_swap(header->sh_addr), eword_swap(header->sh_size));
    fprintf(stdout, " Flags:  %s\n", flags_str);

    if (flags_str)
      free(flags_str);
    if (type_str)
      free(type_str);
}


/* Returns the ASCII representation of the section flags */
char *get_sec_flags(Elf32_Shdr *header)
{
    char *flags;
    Elf32_Word sflags;
    
    sflags = eword_swap(header->sh_flags);
    if ((flags = malloc(64)) == NULL)
      return NULL;
    
    memset(flags, 0x00, 64);
    
    /* process the individual flags */
    if ((sflags & SHF_WRITE) == SHF_WRITE)
      strncpy(flags, "WRITABLE ", 9);
    if ((sflags & SHF_ALLOC) == SHF_ALLOC)
      strncat(flags, "ALLOC ", 6);
    if ((sflags & SHF_EXECINSTR) == SHF_EXECINSTR)
      strncat(flags, "EXECINSTR ", 10);
#ifdef SHF_STRINGS
    if ((sflags & SHF_STRINGS) == SHF_STRINGS)
      strncat(flags, "STRINGS ", 8);
#endif
#ifdef SHF_MERGE
    if ((sflags & SHF_MERGE) == SHF_MERGE)
      strncat(flags, "MERGE ", 6);
#endif
#ifdef SHF_LINK_ORDER
    if ((sflags & SHF_LINK_ORDER) == SHF_LINK_ORDER)
      strncat(flags, "LINK_ORDER ", 11);
#endif
#ifdef SHF_INFO_LINK
    if ((sflags & SHF_INFO_LINK) == SHF_INFO_LINK)
      strncat(flags, "INFO_LINK", 9);
#endif
    
    return flags;
}


/* Returns the ASCII representation of the section type */
char *get_sec_type(Elf32_Shdr *header)
{
    char *sec_type;
    
    if ((sec_type = malloc(32)) == NULL)
      return NULL;
    memset(sec_type, 0x00, 32);
 
    switch (eword_swap(header->sh_type)) {
      case SHT_NULL:
	strncpy(sec_type, "Unused", 31); break;
#ifdef SHT_INIT_ARRAY	
      case SHT_INIT_ARRAY:
	strncpy(sec_type, "Constructor Array", 31); break;
#endif
#ifdef SHT_FINI_ARRAY
      case SHT_FINI_ARRAY:
	strncpy(sec_type, "Destructor Array", 31); break;
#endif
#ifdef SHT_PREINIT_ARRAY
      case SHT_PREINIT_ARRAY:
	strncpy(sec_type, "Pre-constructor Array", 31); break;
#endif
#ifdef SHT_GROUP
      case SHT_GROUP:
	strncpy(sec_type, "Section Group", 31); break;
#endif
#ifdef SHT_SYMTAB_SHNDX
      case SHT_SYMTAB_SHNDX:
	strncpy(sec_type, "Extended Section Indeces", 31); break;
#endif
#ifdef SHT_NUM
      case SHT_NUM:
	strncpy(sec_type, "Number of Defined Types", 31); break;
#endif
	
      case SHT_PROGBITS:
	strncpy(sec_type, "Program Data", 31); break;
      case SHT_SYMTAB:
	strncpy(sec_type, "Symbol Table", 31); break;
      case SHT_STRTAB:
	strncpy(sec_type, "String Table", 31); break;
      case SHT_RELA:
	strncpy(sec_type, "Relocation Entries w/ addends", 31); break;
      case SHT_HASH:
	strncpy(sec_type, "Symbol Hash", 31); break;
      case SHT_DYNAMIC:
	strncpy(sec_type, "Dynamic Linking Info.", 31); break;
      case SHT_NOTE:
	strncpy(sec_type, "Auxiliary Information", 31); break;
      case SHT_NOBITS:
	strncpy(sec_type, "Uninitialised Data (bss)", 31); break;
      case SHT_REL:
	strncpy(sec_type, "Relocation Entries, no addends", 31); break;
      case SHT_DYNSYM:
	strncpy(sec_type, "Dynamic Linker Symbol Table", 31); break;
      default:
	sprintf(sec_type, "Unknown: %d",
		(unsigned char)eword_swap(header->sh_type));
    };
    
    return sec_type;
}
